import json


# loading JSON data
with open("../config.json") as json_config :
    data = json.load(json_config)

print data



# writing/updating JSON data
test_out = data
test_out['grid']['user'] = 'test'

with open("../json_test_out.json","w") as json_out :
    json.dump(test_out,json_out)



from jobs import *
import subprocess
from collections import Counter

user = subprocess.check_output("whoami").strip()

job_list = fetch_jobs(user=user,local=True)

# print job_list
# print len(job_list)

master_jobs = [ job['status'] for job in job_list if job['group'] == 'master']
sub_jobs = [ job['status'] for job in job_list if job['group'] == 'subjob']

# print master_jobs
# print len(master_jobs)

# print sub_jobs
# print len(sub_jobs)

master_counts = Counter(master_jobs)
sub_counts = Counter(sub_jobs)

# print master_counts
# print sub_counts

# dic_master = dict(sub_counts)
# print dic_master

get_status(user,local=True)



import jobs,token
import argparse
import subprocess
import os

def main() :
    ### top-level group (L0)
    parser = argparse.ArgumentParser(description="Welcome to AliGEM - ALICE Grid Enviroment Manager (https://github.com/vpacik/aligem)")
    parser.add_argument("-v","--verbose", help="produce verbose output", action="store_true")
    parser.add_argument("-d","--debug", help="debugging mode (additional printout)", action="store_true")
    parser.add_argument("--version", action="version", version='0.1', help="print current version")
    subparsers = parser.add_subparsers(title="operations",dest="command")

    ### sub-parsers
    # jobs parser (L1)
    parser_jobs = subparsers.add_parser("jobs", help="Grid jobs operations")
    jobs_subparsers = parser_jobs.add_subparsers(dest="job_command")

    # jobs sub-parsers (L2)
    jobs_subparser_status = jobs_subparsers.add_parser("status", description="Print out compact (colored) overview of jobs states:\n['USER'] #M(asterjob)/#S(ubjob) "+jobs.stateclr.ALL+"#ALL"+jobs.stateclr.ENDC+" ("+jobs.stateclr.DONE+"#DONE"+jobs.stateclr.ENDC+"|"+jobs.stateclr.RUNNING+"#RUNNING"+jobs.stateclr.ENDC+"|"+jobs.stateclr.ERROR+"#ERROR"+jobs.stateclr.ENDC+"|"+jobs.stateclr.REST+"#REST"+jobs.stateclr.ENDC+")", help = "print overview of currently registered grid jobs",formatter_class=argparse.RawTextHelpFormatter)
    jobs_subparser_status.add_argument("-u","--user", help="specify USER as CERN username",default=None)
    jobs_subparser_status.add_argument("-f","--full", help="print detailed overview of all jobs states (also invoked by --verbose)", action="store_true")
    jobs_subparser_status.add_argument("--only-positive", help="print only states with at least 1 (sub)job", action="store_true")

    jobs_subparser_kill = jobs_subparsers.add_parser("kill", help = "kill grid (sub)job(s) in DONE state")
    jobs_subparser_kill.add_argument("-r","--resub", help="resubmit failed jobs prior to sequential killing", action="store_true",dest="kill_resub")
    jobs_subparser_kill.add_argument("-A","--all", help="kill ALL registered jobs (independent of state)", action="store_true",dest="kill_all")
    # jobs_subparser_kill.add_argument("-u","--user", help="specify USER as CERN username")

    jobs_subparser_resubmit = jobs_subparsers.add_parser("resub", help = "re-submit all grid job(s) in ERROR, EXPIRED or ZOMBIE state")

    # token parser (L1)
    parser_token = subparsers.add_parser("token", help="AliEn token operations")
    token_subparsers = parser_token.add_subparsers(dest="token_command")
    token_subparser_init = token_subparsers.add_parser("init", help="Initialize new token")
    token_subparser_destroy = token_subparsers.add_parser("destroy", help="Destroy current token")
    token_subparser_info = token_subparsers.add_parser("info", help="List token information")

    args = parser.parse_args()
    args = vars(args) # make an dictionary out of namespace

    debug = args['debug']
    verbose = args['verbose']

    if debug :
        print "=== Arguments ================================="
        print args
        print "==============================================="

    # check if alien is within the $PATH
    if not check_alien(debug=debug) :
        print "AliEn not found in $PATH, please load alienv !"
        return

    if 'command' not in args :
        print "Something went wrong, no 'command' argument parsed."
        return

    if args['command'] == 'jobs' :
        if debug : print "inside jobs"

        if 'job_command' not in args :
            print "Something went wrong, no 'job_command' argument parsed."
            return

        # check for valid token (if not found, token-init)
        if (token.check() == False) :
            print "No valid token found. Initializiting new token!"
            token.init()

        local_user = jobs.exec_alien_cmd("alien_whoami")['output'].strip()

        if args['job_command'] == 'status' :
            if debug : print "inside status"

            if args['user'] is not None :
                local_user = args['user']

            if debug : print local_user
            if not verbose and args['full'] is True : verbose = True

            jobs.get_status(local_user, debug=debug, verbose=verbose,only_positive=args['only_positive'])

        if args['job_command'] == 'kill' :
            if debug : print "inside kill"
            # user = args.user

            if args['kill_all'] == True :
                if debug : print "kill all is ON!"
                jobs.kill_all(local_user, debug=debug)
            else :
                if debug : print "kill done only"
                jobs.kill_done(local_user,resub=args['kill_resub'], debug=debug)


        if args['job_command'] == 'resub' :
            if debug : print "inside resubmit"
            # user = args.user
            jobs.resubmit(local_user,debug=debug)


    if args['command'] == 'token' :
        if 'token_command' not in args :
            print "Something went wrong, no 'token_command' argument parsed."
            return

        if args['token_command'] == "init" :
            if not token.check() :
                token.init()
            else :
                print "Valid token already exists! Destroy it first!"

        if args['token_command'] == "destroy" :
            token.destroy()

        if args['token_command'] == "info" :
            token.info()
    return

def check_alien(debug=False) :
    """
    Check if AliEn commands are available (in $PATH)
    """
    if debug : print "check_alien()"
    path = os.getenv('PATH')

    if "/AliEn-Runtime/" not in path :
        return False

    return True



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='aligem',
      version='1.1.0',
      description='ALICE Grid Enviroment Manager',
      long_description=readme(),
      keywords='aligem ALICE CERN high energy physics',
      url='http://github.com/vpacik/aligem',
      author='Vojtech Pacik',
      author_email='vojtech.pacik@cern.ch',
      license='GPL-3.0',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'Natural Language :: English',
      ],
      packages=['aligem'],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      entry_points = {
          'console_scripts': ['aligem=aligem.command_line:main'],
      },
      include_package_data=True,
      zip_safe=False)



# Jobs handling suite

import subprocess
import argparse
from collections import Counter

class stateclr :
    """ Object with defined output colors for job states for terminal (ASCII) """
    ALL = '\033[1;35m'
    DONE = '\033[92m'
    RUNNING = '\033[36m'
    ERROR = '\033[91m'
    REST = '\033[93m'
    ENDC = '\033[0m'

def main() :
    print "Welcome to AliGEM | Jobs"

def exec_alien_cmd(process = [], verbose=False) :
    """
    Wrapper function to execude terminal commands within python based on subprocess package.
    Example of input process list ['alien_top','-all_status','-user',str(user)]
     - first element is the command
     - following arguments is parameter (& values) sequence
    TODO: ATM it is handled by try-except clause, a Popen might be more suitable (and stable?) solution
    """

    if not process :
        print 'Process list is empty. Nothing to execute!'
        return

    try :
        output = subprocess.check_output(process)
        return { "cmd" : process, "returncode" : 0, "output" : output }

    except subprocess.CalledProcessError as error :
        if(verbose) :
            print("CalledProcessError exception caught!")
            print("Error return code:\n" + str(error.returncode))
            print("Error cmd:\n" + str(error.cmd))
            print("Error output:\n" + str(error.output))

        return { "cmd" : error.cmd, "returncode" : error.returncode, "output" : error.output }

    except OSError as error :
        print "OSError expection caught! Unknown process '" + str(process) + "'"
        if verbose : print error

    except ValueError as error :
        print "ValueError expection caught! Popen (likely) invoked with invalid arguments"
        if verbose : print error

    return { "cmd" : str(process), "returncode": None, "output" : str(error) }

def fetch_jobs(user,verbose=False,debug=False) :
    """
    Returns list of job dicts from Grid query for a single user
    """
    job_string = exec_alien_cmd(['alien_top','-all_status','-user',str(user)],verbose=verbose)['output']

    # stripping useless characters within jobs output
    job_string = job_string.replace(" ","") # removing whitespaces
    job_string = job_string.rstrip("\n\n") # stripping last \n\n to avoid empty entries
    jobs = job_string.split("\n\n")

    list_jobs = []
    for job in jobs :
        processed_job = process_single_job(job)
        if validate_single_job(processed_job,debug=debug) : list_jobs.append(processed_job)
    return list_jobs

def process_single_job(job='') :
    """ Returns a dict with a data from single job"""
    job = job.split("\t")

    if len(job) < 4 :
        return None

    if job[4].startswith("pcapiserv") :
        job_type = 'master'
    else :
        job_type = 'subjob'

    single_job = { 'id':job[0], 'group':job_type, 'status':job[1], 'user':job[2], 'script':job[3], 'server':job[4] }
    return single_job

def validate_single_job(job,debug=False) :
    """ Returns True if job is valid, False otherwise """
    isOK = True

    if job is None :
        if debug : print "invalid job"
        return False

    if not job['id'].isdigit() :
        isOK = False
        if debug : print "job id '%s' not validated" % str(job['id'])

    if ( (job['group'] == 'master') and (not job['server'].startswith('pcapiserv')) ) :
        isOK = False
        if debug : print "job group '%s' not validated" % job['group']

    if (job['group'] == 'subjob') and (not job['server'].startswith('aliendb')) :
        isOK = False
        if debug : print "job group '%s' not validated" % job['group']

    if debug and (isOK == False) : print(job)
    return isOK

def get_status(user='vpacik',verbose=False,debug=False,only_positive=False) :
    """
    Fetching jobs from Grid servers, sorting them according to their status and prints brief overview
    """

    if user == None :
        print 'User not specified. This might take long time. Aborted!'
        return

    # Works well except for states with starts with something; ie. ERROR_*, DONE_*

    jobs = fetch_jobs(str(user),debug=debug)
    if len(jobs) == 0 :
        print "No (validated) jobs found."
        return

    master = [ job['status'] for job in jobs if job['group'] == 'master' ]
    num_master = len(master)
    counts_master = Counter(master)

    sub_jobs = [ job['status'] for job in jobs if job['group'] == 'subjob' ]
    num_subjob = len(sub_jobs)
    counts_subjob = Counter(sub_jobs)

    # getting general counts of finer split elements (such as ERROR_* and DONE_*)
    for key,value in counts_master.iteritems() :
        if key.startswith("ERROR") : counts_master += Counter({'ERROR_ALL' : value})
        if key.startswith("DONE") : counts_master += Counter({'DONE_ALL' : value})

    for key,value in counts_subjob.iteritems() :
        if key.startswith("ERROR") : counts_subjob += Counter({'ERROR_ALL' : value})
        if key.startswith("DONE") : counts_subjob += Counter({'DONE_ALL' : value})

    # printing short (one-line) version of status
    if not verbose :
        # print " M %d (D%d|R%d|E%d) S %d (D%d|R%d|E%d) " % (num_master,counts_master['DONE_ALL'],counts_master['RUNNING'],counts_master['ERROR_ALL'], num_subjob, counts_subjob['DONE_ALL'],counts_subjob['RUNNING'],counts_subjob['ERROR_ALL'])
        out = "--- Jobs status ['%s'] " % str(user)
        out += "#M "+stateclr.ALL + str(num_master) + stateclr.ENDC +" ("
        out += stateclr.DONE + str(counts_master['DONE_ALL']) + stateclr.ENDC + "|"
        out += stateclr.RUNNING + str(counts_master['RUNNING']) + stateclr.ENDC + "|"
        out += stateclr.ERROR + str(counts_master['ERROR_ALL']) + stateclr.ENDC + "|"
        out += stateclr.REST + str(num_master - counts_master['DONE_ALL'] - counts_master['RUNNING'] - counts_master['ERROR_ALL']) + stateclr.ENDC+ ")"
        out += " #S "+stateclr.ALL + str(num_subjob) + stateclr.ENDC +" ("
        out += stateclr.DONE + str(counts_subjob['DONE_ALL']) + stateclr.ENDC + "|"
        out += stateclr.RUNNING + str(counts_subjob['RUNNING']) + stateclr.ENDC + "|"
        out += stateclr.ERROR + str(counts_subjob['ERROR_ALL']) + stateclr.ENDC + "|"
        out += stateclr.REST + str(num_subjob - counts_subjob['DONE_ALL'] - counts_subjob['RUNNING'] - counts_subjob['ERROR_ALL']) + stateclr.ENDC+")"

        print out
        return

    # lists of job states used for ordered printing (NB: counts.keys() could be used instead, however it will be "randomly" ordered)
    states_master = [ "DONE_ALL", "SPLIT", "INSERTING", "RUNNING", "SAVING", "SAVED", "ERROR_ALL", "ZOMBIE", "REST"]
    states_subjob = [ "DONE_ALL", "WAITING", "ASSIGNED", "STARTED", "RUNNING", "SAVING", "SAVED", "ERROR_ALL", "EXPIRED", "ZOMBIE", "REST"]

    # estimating REST state (one not listed (and thus printed) in above lists)
    num_rest = num_master
    for key, value in counts_master.iteritems() :
        if key in states_master : num_rest -= value
    counts_master.update({'REST' : num_rest})

    num_rest = num_subjob
    for key, value in counts_subjob.iteritems() :
        if key in states_subjob : num_rest -= value
    counts_subjob.update({'REST' : num_rest})

    # Printing status output to the user
    def printStatusLine(label, num, num_all, only_positive=only_positive) :
        perc = -1.0
        if num_all > 0.0 : perc = 100*float(num)/num_all
        if only_positive and num == 0 : return

        print '%10s: %5d %6.1f%%' % (label, num, perc)
        return

    print '######################################'
    print '  Jobs status for user "%s"' % user
    print '======= Masterjobs (%d) ==============' % num_master
    for key in states_master :
        printStatusLine(key, counts_master[key], num_master)

    if not only_positive or (only_positive and num_subjob > 0) :
        print '======= Subjobs (%d) =================' % num_subjob
        for key in states_subjob :
            printStatusLine(key, counts_subjob[key], num_subjob)

    print '######################################'

    if debug :
        print "Master :",; print counts_master
        print "Subjob :",; print counts_subjob
    return

def kill_job_id(job_id, verbose=False) :
    """ Kill a single job based on input id """
    print exec_alien_cmd(['alien_kill',str(job_id)],verbose=verbose)['output']

def kill_jobs(jobs_list,verbose=False,debug=True) :
    """ Kill ALL jobs based on provided list with (filtered) jobs. """
    if not jobs_list :
        print 'List of jobs is empty. Nothing to kill'
        return

    for job in jobs_list :
        job_id = job['id']

        if debug :
            print 'Job ID to kill %s (would be killed, running in debug mode)' % (job_id)
            print job
        else :
            kill_job_id(job_id,verbose=verbose)
    return

def filter_jobs(list_jobs, group=None, status=None, server=None, user=None, verbose=False) :
    """ Returns a list with jobs passing filtering criteria """

    if not list_jobs : # check if list is empty
        if verbose : print 'Input job list is empty. Nothing to filter'
        return []

    filtered_jobs = list_jobs

    if user != None :
        filtered_jobs = [ job for job in filtered_jobs if job['user'] == str(user) ]

    if group != None :
        filtered_jobs = [ job for job in filtered_jobs if job['group'] == str(group) ]

    if server != None :
        filtered_jobs = [ job for job in filtered_jobs if job['server'].startswith(str(server)) ]

    if status != None :
        filtered_jobs = [ job for job in filtered_jobs if job['status'].startswith(str(status)) ]

    return filtered_jobs

def kill_done(user, resub=False, verbose=False, debug=False) :
    """
    Kill jobs in DONE state. If resub=True, jobs in error states are resubmitted first, and then master and subjobs are
    killed sequentially. If resub=False, only subjobs are killed avoiding killing master jobs with failed subjobs (see NB).
    NB: Do not kill master jobs first (ala resubmit()) since master job is in done state even if subjobs are in ERROR!
    """

    if resub == True :
        groups = ["master","subjob"]
        resubmit(user=user, verbose=verbose, debug=debug)
        if debug : print "Resubmitted"
    else :
        groups = ["subjob"]

    for group in groups :
        jobs_list = fetch_jobs(user,verbose=verbose,debug=debug)
        filtered = filter_jobs(jobs_list,group=group, status="DONE",verbose=verbose)

        if verbose : print "Number of %s jobs to be deleted: %d " % str(group),len(filtered)
        kill_jobs(filtered, debug=debug)
    return

def kill_all(user, verbose=False, debug=False) :
    """ Kill ALL jobs independent of state """
    filtered = filter_jobs(fetch_jobs(user),group="master",verbose=verbose)
    if verbose : print "Number of jobs to be deleted: %d " % len(filtered)
    kill_jobs(filtered, debug=debug)
    return

def resubmit(user, verbose=False, debug=False) :
    """ Resubmit all jobs in error (ERROR, EXPIRED, ZOMBIE) states. This is done sequentially: first MASTER then SUBJOB jobs"""
    groups = ["master","subjob"]

    for group in groups :
        jobs_list = fetch_jobs(user,verbose=verbose,debug=debug)
        jobs_group = filter_jobs(jobs_list,group=group,verbose=verbose)

        filtered = []
        filtered.extend(filter_jobs(jobs_group,status="ERROR",verbose=verbose))
        filtered.extend(filter_jobs(jobs_group,status="EXPIRED",verbose=verbose))
        filtered.extend(filter_jobs(jobs_group,status="ZOMBIE",verbose=verbose))

        if len(filtered) == 0 :
            print "No %s jobs to resubmit!" % str(group)
            continue

        for job in filtered :
            job_id = job['id']

            if debug :
                print "Resubmitting job id %s" % str(job_id)
            print exec_alien_cmd(['alien_resubmit',str(job_id)],verbose=verbose)['output']

    return

# ==============================================================================



# AliEn token operations handler

import jobs

def main() :
    print "Token.main() says hi"

def init(debug=False, verbose=False) :
    if debug : print "init inside"
    output = jobs.exec_alien_cmd('alien-token-init',verbose)

    if output['returncode'] == 0 :
        print "Token succesfully initialized!"
    else :
        print "Token NOT initialized!"
        if verbose :
            print output['output'].strip()

def info(debug=False, verbose=False) :
    if debug : print "token.info()"
    output = jobs.exec_alien_cmd('alien-token-info',verbose)
    print output['output'].strip()

def destroy(debug=False,verbose=False) :
    if debug : print "destroy inside"
    output = jobs.exec_alien_cmd('alien-token-destroy',verbose)

    if output['returncode'] == 0 :
        print "Token succesfully destroyed!"
    else :
        print "Token not destoyed"
        if verbose :
            print output['output'].strip()

def check(debug=False,verbose=False) :
    token = jobs.exec_alien_cmd('alien-token-info',verbose)

    if token['returncode'] == 1 :
        # not succesfull (no token)
        if verbose : print "Token not validated!"
        if debug : print token
        return False
    else :
        # info succesfull but checking token content
        item = token['output']
        item = item.replace(" ","")
        item = item.split("\n")
        # print item

        # checking host
        if item[0].split(":")[1] == "" :
            print item[0].split(":")[1]
            return False

        # checking Password
        if item[4].split(":")[1] == "" :
            print item[4].split(":")[1]
            return False

        # checking Nonce
        if item[5].split(":")[1] == "" :
            print item[5].split(":")[1]
            return False

        # checking SID
        if item[6].split(":")[1] == "" :
            print item[6].split(":")[1]
            return False

    return True
